#! /usr/bin/env/python3
# -*- coding=utf-8 -*-
import time
import logging
import nacos
import uvicorn
import socket
import redis
from fastapi import FastAPI, Request
from fastapi.exceptions import RequestValidationError
from fastapi.responses import JSONResponse
from nacos import NacosException
from typing import Optional

from shared.conf import PORT, RPC_LOCATION, VERSION, NACOS_ENABLE
from shared.exception_handler import NacosContext
from shared.logging_util import logger as fine_ai_logger
from shared.nacos_helpers import register_nacos_service, deregister_nacos_service
from shared.rpc_toolkit import RpcPayload, RpcResponse
from chat2search.app import rpc as chat2search_rpc, startup as chat2search_startup, shutdown as chat2search_shutdown
from chat2search.utils.logging_util import logger as chat2search_logger
from fine_agent.app import router as fine_agent_router
from fine_agent.utils.logging_util import logger as fine_agent_logger
from ai_core_suite.app import router as ai_core_suite_router
from ai_core_suite.utils.logging_util import logger as ai_core_suite_logger
from shared.status_codes import *


app = FastAPI()
app.include_router(fine_agent_router)
app.include_router(ai_core_suite_router)


nacos_client: nacos.NacosClient = None
group_list = ("decision", "bi")
credential_dict = dict()


def update_credential_callback(config):
    """
    Nacos 配置监听器触发的回调函数，入参 config 是一个字典，其 key 为：
    'data_id', 'group', 'namespace', 'raw_content', 'content'
    """
    chat2search_logger.info(f"Credential callback")
    global credential_dict
    credential_dict[config["group"]] = config["content"]


def add_nacos_config_watchers(group_list):
    """监听 Nacos 上数据中心和平台的 credential"""
    data_id = "credential"
    for group in group_list:
        credential_dict[group] = nacos_client.get_config(data_id, group, timeout=10)
        nacos_client.add_config_watchers(data_id, group, [update_credential_callback])


def valid_credential(rpc_payload: RpcPayload):
    group = rpc_payload.sourceLocation
    data_id = "credential"
    if not rpc_payload.credential:
        return False
    if group not in credential_dict:
        add_nacos_config_watchers([group])
        fine_ai_logger.info(f"新增 credential 配置监听器，sourceLocation: {group}")
    if credential_dict[group] is None:
        # nacos 配置监听疑似有 bug，此处添加兜底机制，主动获取 credential
        credential_dict[group] = nacos_client.get_config(data_id, group, timeout=10)
        if credential_dict[group] is None:
            raise NacosException(f"Not found data_id '{data_id}' of group '{group}' in namespace '{nacos_client.namespace}'")
    if credential_dict[group] == rpc_payload.credential:
        return True
    else:
        return False


def get_logger(rpc_payload):
    if rpc_payload.serviceName == "chat2search":
        return chat2search_logger
    elif rpc_payload.serviceName == "fine-agent":
        return fine_agent_logger
    elif rpc_payload.serviceName == "ai-core-suite":
        return ai_core_suite_logger
    else:
        return fine_ai_logger


@app.middleware("http")
async def handle_response(request, call_next):
    response = await call_next(request)
    response.headers["Content-Type"] = "application/json; charset=UTF-8"
    return response


@app.exception_handler(RequestValidationError)
def validation_exception_handler(request: Request, exc: RequestValidationError):
    fine_ai_logger.warning(f"RequestValidationError: {exc.errors()}")
    return JSONResponse(
        status_code=200,
        content=RpcResponse(error_code=CODE_BAD_REQUEST, error_msg=f"RequestValidationError: {exc.errors()}"),
    )


@app.on_event("startup")
async def startup_event():
    fine_ai_logger.info(f"------------ Version {VERSION} ------------")
    await chat2search_startup()
    if NACOS_ENABLE:
        global nacos_client
        fine_ai_logger.info(f"启用 NACOS 服务发现")
        with NacosContext():
            nacos_client = register_nacos_service()
            add_nacos_config_watchers(group_list)
    else:
        fine_ai_logger.info(f"NACOS 未启用，不支持服务发现")
    fine_ai_logger.info("fine_ai starts successfully")
    fine_ai_logger.info(f"------------ Version {VERSION} ------------")


@app.on_event("shutdown")
async def shutdown_service():
    try:
        await chat2search_shutdown()
        if NACOS_ENABLE:
            data_id = "credential"
            for group in group_list:
                nacos_client.remove_config_watcher(data_id, group, update_credential_callback, True)
            deregister_nacos_service(nacos_client)
    except nacos.NacosException:
        fine_ai_logger.warning(f"注销 Nacos 服务失败")
    except Exception as e:
        fine_ai_logger.exception(e)
        fine_ai_logger.warning(f"服务关闭时出现未知错误")
    fine_ai_logger.info(f"fine_ai 服务已关闭")


@app.get(f"/webroot/{RPC_LOCATION}/alive")
def is_alive():
    return True


@app.get("/version")
def is_alive():
    return VERSION


# @app.post(f"/webroot/{RPC_LOCATION}/rpc")
async def rpc(rpc_payload: RpcPayload, tenantId: Optional[str] = "default"):
    """根据 rpc_payload.serviceName 将请求体分发给各项目"""
    logger = get_logger(rpc_payload)
    try:
        if not NACOS_ENABLE:
            fine_ai_logger.error(f"NACOS 服务未启用，无法处理 rpc 请求")
            return RpcResponse(error_code=CODE_NACOS_ERROR, error_msg="NACOS 服务未启用，无法处理 rpc 请求")
        if not valid_credential(rpc_payload):
            fine_ai_logger.warning(f"Invalid credential from sourceLocation '{rpc_payload.sourceLocation}'")
            return RpcResponse(error_code=CODE_INVALID_CREDENTIAL, error_msg="Invalid credential")
        if rpc_payload.serviceName == "chat2search":
            return await chat2search_rpc(rpc_payload)
        # elif rpc_payload.serviceName == "llm-adapter":
        #     pass
        # elif rpc_payload.serviceName == "ai-core-suite":
        #     pass
        else:
            logger.warning(f"Unknown payload.serviceName {rpc_payload.serviceName}")
            return RpcResponse(error_code=CODE_BAD_REQUEST, error_msg="Unknown payload.serviceName")
    except NacosException as e:
        logger.warning(f"Nacos Error: {str(e)}")
        return RpcResponse(error_code=CODE_NACOS_ERROR, error_msg=f"Nacos Error: {str(e)}")
    except Exception as e:
        logger.exception(e)
        return RpcResponse(error_code=CODE_AI_ERROR, error_msg="AI Error")


@app.put("/loglevel/{level}")
def set_log_level(level: str):
    try:
        level = level.strip().lower()
        if level == "critical":
            log_level = logging.CRITICAL
        elif level == "fatal":
            log_level = logging.FATAL
        elif level == "error":
            log_level = logging.ERROR
        elif level == "warning":
            log_level = logging.WARNING
        elif level == "info":
            log_level = logging.INFO
        elif level == "debug":
            log_level = logging.DEBUG
        else:
            return False
        for logger in [chat2search_logger, fine_agent_logger, ai_core_suite_logger, fine_ai_logger]:
            logger.setLevel(log_level)
        return True
    except Exception as e:
        fine_ai_logger.exception(e)
        return False


def get_local_ip():
    skt = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    skt.connect(("8.8.8.8", 80))
    return skt.getsockname()[0]


if __name__ == "__main__":
    uvicorn.run(app='app:app', host=get_local_ip(), port=PORT, reload=False, timeout_keep_alive=300)
